local AceComm = LibStub("AceComm-3.0")
local Serializer = LibStub("AceSerializer-3.0")
local Compressor = LibStub("LibCompress")

local Icon = LibStub("LibDBIcon-1.0")
local LDBobject = LibStub("LibDataBroker-1.1"):NewDataObject("BossList", {
	type = "launcher",
	icon = "Interface\\Buttons\\UI-GroupLoot-Dice-Up",
  OnTooltipShow = function(tooltip) 
    tooltip:AddLine("BossList")
    tooltip:AddLine("Click to open the overview of needs")
    tooltip:AddLine("Right click to open the settings") end,
  label = "BossList",
	OnClick = function() 
    if GetMouseButtonClicked() ==  "RightButton" then
      InterfaceOptionsFrame_OpenToCategory("BossList");
    else
      PlaySound("igMainMenuOptionCheckBoxOn")
      BossList.frame:Show()
    end end,
})

local Encoder = Compressor:GetAddonEncodeTable()
local AceConfig = LibStub("AceConfig-3.0")
local AceDB = LibStub("AceDB-3.0")
local DB

-- Message prefixes
local FULL_MSG_PREFIX = "BL_FULL_MSG"
local SMALL_MSG_PREFIX = "BL_SMALL_MSG"

local ADDON_PREFIX = "BossList"
local REQUEST_UPDATER = "REQ_UPDATER"
local REQUEST_SYNC = "REQ_SYNC"
local AVAILABLE = "AVAILABLE"
local ENTRY_MESSAGE = "(%d+)%-(%d+)%-(%d+)%-(%a+)%-(%d+)%-(%d+)"

local REVISION = 1

local requestedUpdater = false

local entries
local frame = CreateFrame("Frame")
local DisplayEncounter_Orig
local currentInstanceID
local currentEncounterID
local dropDown

local desires = {"Need", "Minor", "Off-spec", "Not needed"}
local desireIcons = {"N", "M", "O", " "}
local desireColors = {{r=0, g=1, b=0}, {r=1, g=1, b=0}, {r=1, g=0.5, b=0}}

local defaults = {
  char = {
    minimap = {
      hide = false,
    }
  },
  global = {
    hideNotNeeded = true
  }
}


function BossList_DEBUG_PrintEntriesIDs(table)
  print("Printing contents of table")
  local table = table or BossList_entries
  print(table)
  for instance, difficulties in pairs(table) do             -- Instances
    for difficulty, encounters in pairs(difficulties) do    -- Difficulties
      for encounter, desires in pairs(encounters) do        -- Encounters
        for desire, characters in pairs(desires) do         -- Desires
          for character, lastUpdate in pairs(characters) do -- Characters
            local msg = string.format("%d-%d-%d-%s-%d-%d", instance, difficulty, encounter, character, desire, lastUpdate)
            print(msg)
          end
        end
      end
    end
  end
end

function BossList_DEBUG_PrintEntries(table)
  print("Printing contents of table")
  local table = table or BossList_entries
  print(table)
  for instance, difficulties in pairs(table) do             -- Instances
    instanceName = EJ_GetInstanceInfo(instance) or "Unknown"
    for difficulty, encounters in pairs(difficulties) do    -- Difficulties
      for encounter, desires in pairs(encounters) do        -- Encounters
        encounterName = EJ_GetEncounterInfo(encounter) or "Unknown"
        for desire, characters in pairs(desires) do         -- Desires
          for character, lastUpdate in pairs(characters) do -- Characters
            lastUpdateTime = date("%d/%m/%y %H:%M:%S", lastUpdate) or "Unknown"
            local msg = string.format("%s-%d-%s-%s-%d-%s", instanceName, difficulty, encounterName, character, desire, lastUpdateTime)
            print(msg)
          end
        end
      end
    end
  end
end

local playerName = GetUnitName("player", false)

-- Extract the difficulty from the EncounterJournal link
local function GetDifficulty(link)
  local _, _, difficulty = link:find("%d+:%d+:(%d+)|")
  return tonumber(difficulty)
end

-- Might be used to catch changes between versions
local function CheckRevision()
  if DB.revision == nil or DB.revision < REVISION then
    --BossList_entries = {}
    --print("BossList: Wiped local database due to update")
    DB.revision = REVISION
    
  end
end

-- Return the entry if it exists, otherwise return nil
local function getEntry(table, instanceID, difficulty, encounterID, characterName)
  if table[instanceID] == nil or table[instanceID][difficulty] == nil or  -- Check if the encounter table exists
     table[instanceID][difficulty][encounterID] == nil then
    return nil
  end
  
  -- Check all desires for an entry
  for desire, characters in pairs(table[instanceID][difficulty][encounterID]) do
    if characters[characterName] ~= nil then
      return desire, characters[characterName]
    end
  end
  
  -- Nothing found
  return nil
end

-- Add entry into table
local function addEntry(table, instanceID, difficulty, encounterID, characterName, desirability, lastUpdate)
  if table[instanceID] == nil then table[instanceID] = {} end
  local instanceTable = table[instanceID]
  if instanceTable[difficulty] == nil then instanceTable[difficulty] = {} end
  local difficultyTable = instanceTable[difficulty]
  if difficultyTable[encounterID] == nil then difficultyTable[encounterID] = {} end
  local encounterTable = difficultyTable[encounterID]
  if encounterTable[desirability] == nil then encounterTable[desirability] = {} end
  local desireTable = encounterTable[desirability]
  
  desireTable[characterName] = lastUpdate or time()
end

-- Send table using prefix
local function SendTable(prefix, table, target)
  table.revision = REVISION
  
  local data = Serializer:Serialize(table)
  local compData = Compressor:CompressHuffman(data)
  local encData = Encoder:Encode(compData)
  
  -- If a target is provided, whisper the message. Otherwise broadcast to guild.
  if target == nil then
    AceComm:SendCommMessage(prefix, encData, "GUILD")
  else
    AceComm:SendCommMessage(prefix, encData, "WHISPER", target)
  end
  
  table.revision = nil
end

-- Compare all the entries in table to the local data. If newer or local is missing add it.
local function CompareData(table)
  for instance, difficulties in pairs(table) do             -- Instances
    for difficulty, encounters in pairs(difficulties) do    -- Difficulties
      for encounter, desires in pairs(encounters) do        -- Encounters
        for desire, characters in pairs(desires) do         -- Desires
          for character, lastUpdate in pairs(characters) do -- Characters
            local localDesire, localLastUpdate = getEntry(BossList_entries, instance, difficulty, encounter, character)
            if localDesire == nil or localLastUpdate <= lastUpdate then
              -- No local entry or received entry is newer, update local data.
              BossList.frame.addEntry(instance, difficulty, encounter, character, desire, lastUpdate)
            end
          end
        end
      end
    end
  end
end

-- Compare all the entries in table to the local data. If newer or local is missing add it. Any local entries missing
-- in table are sent to the guild.
local function MergeData(table)
  local updatesTable = {}
  for instance, difficulties in pairs(BossList_entries) do  -- Instances
    for difficulty, encounters in pairs(difficulties) do    -- Difficulties
      for encounter, desires in pairs(encounters) do        -- Encounters
        for desire, characters in pairs(desires) do         -- Desires
          for character, lastUpdate in pairs(characters) do -- Characters
            local recvDesire, recvLastUpdate = getEntry(table, instance, difficulty, encounter, character)
            if recvDesire == nil then
              -- Local entry exists but not contained in received table, broadcast
              addEntry(updatesTable, instance, difficulty, encounter, character, desire, lastUpdate)
            else
              if recvLastUpdate < lastUpdate then
                -- Local entry is newer, broadcast
                addEntry(updatesTable, instance, difficulty, encounter, character, desire, lastUpdate)
              else
                -- Received entry is newer, update local entry
                BossList.frame.addEntry(instance, difficulty, encounter, character, recvDesire, recvLastUpdate)
              end
              table[instance][difficulty][encounter][recvDesire][character] = nil
            end
          end
        end
      end
    end
  end
  
  -- Send the updates to the guild
  if next(updatesTable) == nil then
    SendTable(SMALL_MSG_PREFIX, updatesTable)
  end
  
  -- Check for any entries not in local table
  for instance, difficulties in pairs(table) do             -- Instances
    for difficulty, encounters in pairs(difficulties) do    -- Difficulties
      for encounter, desires in pairs(encounters) do        -- Encounters
        for desire, characters in pairs(desires) do         -- Desires
          for character, lastUpdate in pairs(characters) do -- Characters
            -- Insert in local table
            BossList.frame.addEntry(instance, difficulty, encounter, character, desire, lastUpdate)
          end
        end
      end
    end
  end
end

local function UpdateHandler(prefix, text, distribution, sender)
  if sender == playerName then
    return
  end
  
  --print("Received update - " .. prefix .. " from " .. sender)
  -- Parse the message
  local compData = Encoder:Decode(text)
  local data, msg = Compressor:Decompress(compData)
  local succes, table = Serializer:Deserialize(data)
  
  if data == nil then
    print("BossList: Failed to decompress - " .. msg)
    print("BossList: Please tell Saberos")
    return
  end
  
  if not succes then
    print("BossList: Failed to deserialize - " .. table)
    print("BossList: Please tell Saberos")
    return
  end
  
  if table.revision == nil or table.revision < REVISION then
    print("BossList: Received message from older version from " .. sender .. ". Please tell him/her to update.")
    return
  else
    table.revision = nil
  end
  
  -- Update local data
  if prefix == FULL_MSG_PREFIX then
    MergeData(table)
  elseif prefix == SMALL_MSG_PREFIX then
    CompareData(table)
  end  
end

-- Send a single entry to the guild
local function SendUpdate(instance, difficulty, encounter, character, desirability, lastUpdate)
  local table = {}
  addEntry(table, instance, difficulty, encounter, character, desirability, lastUpdate)
  SendTable(SMALL_MSG_PREFIX, table)
end

local function RequestSync(player)
  requestedUpdater = false
  AceComm:SendCommMessage(ADDON_PREFIX, REQUEST_SYNC, "WHISPER", player)
  --print(string.format("Send REQUEST_SYNC to %s", player))
end

local function RequestUpdater()
  requestedUpdater = true
  AceComm:SendCommMessage(ADDON_PREFIX, REQUEST_UPDATER, "GUILD")
  --print("Send REQUEST_UPDATER")
end

local function SendAvailable(player)
  AceComm:SendCommMessage(ADDON_PREFIX, AVAILABLE, "WHISPER", player)
  --print("Send AVAILABLE")
end

local function SyncPlayer(player)
  SendTable(FULL_MSG_PREFIX, BossList_entries, player)
  --print("Send BossList_entries")
end

local function MessageHandler(prefix, text, distribution, sender)  
  if sender == playerName then
    return
  end
  
  --print("Received " .. text .. " from " .. sender)
  if text == REQUEST_UPDATER then
    SendAvailable(sender)
  elseif text == REQUEST_SYNC then
    SyncPlayer(sender)
  elseif text == AVAILABLE then
    if requestedUpdater then
      RequestSync(sender)
    end
  end
end

-- Deletes entries older than 3 months
local function CheckTable()
  for instance, difficulties in pairs(BossList_entries) do -- Instances
    for difficulty, encounters in pairs(difficulties) do -- Difficulties
      for encounter, desires in pairs(encounters) do -- Encounters
        for desire, characters in pairs(desires) do -- Desires
          for character, lastUpdate in pairs(characters) do -- characters
            if time() - lastUpdate > 7776000 then -- Older than 3 months
              
              BossList_entries[instance][difficulty][encounter][desire][character] = nil
            end
          end
          if next(characters) == nil then
            desires[desire] = nil
          end            
        end
        if next(desires) == nil then
          encounters[encounter] = nil
        end
      end
      if next(encounters) == nil then
        difficulties[encounter] = nil
      end
    end
    if next(difficulties) == nil then
      BossList_entries[instance] = nil
    end
  end
end

local function EventHandler(self, event, ...)
  if event=="ADDON_LOADED" then
    local name = ...
    if name == "BossList" then   
      -- Create options
      DB = AceDB:New("BossList_DB", defaults)
      BossList.DB = DB
      
      if BossList_entries == nil then
        --print("BossList: First time loaded, initialized BossList_entries")
        BossList_entries = {}
      end
      CheckRevision()
      CheckTable() -- Check for old entries

      -- Attempt to request an updater
      requestedUpdater = true
      frame.tryCounter = 5    -- Number of attempts before giving up
      frame.OnUpdate = function(self, elapsed) 
        self.lastUpdate = (self.lastUpdate or 5) + elapsed -- Do the first check after 5 seconds, then every 10.
        if self.lastUpdate >= 10 then
          self.lastUpdate = 0
          if requestedUpdater and self.tryCounter > 0 then
            if IsInGuild() then
              RequestUpdater()
            end
            self.tryCounter = self.tryCounter - 1
          else
            requestedUpdater = false
            self.tryCounter = 0            
            frame:SetScript("OnUpdate", nil)
          end
        end
      end
      
      frame:SetScript("OnUpdate",frame.OnUpdate)
      frame:Show()
      
      -- Create minimap button
      Icon:Register("BossList", LDBobject, DB.char.minimap)     
      
    
    elseif name == "Blizzard_EncounterJournal" then 
      DisplayEncounter_Orig = EncounterJournal_DisplayEncounter;
      
      Init()
      frame:UnregisterEvent("ADDON_LOADED")
    end
  end
end

function HideButtons()
  local _G = _G
  local counter = 1
  local BossEntry = _G["EncounterJournalBossButton"..counter]
  
  while BossEntry ~= nil do
    local bossButton = BossEntry.bossButton
    if bossButton ~= nil then
      bossButton:Hide()
    end

    counter = counter + 1
    BossEntry = _G["EncounterJournalBossButton"..counter]
  end
end

local function BossList_Menu_OnClick(self, desire, arg2, checked)  
  PlaySound("igMainMenuOptionCheckBoxOn")
  if not(IsInGuild()) then
    print("BossList: You are not in a guild, you can't change your choices")
    return
  end

  dropDown.button:SetText(desireIcons[desire])
  local color = desireColors[desire]
  if color ~= nil then
    dropDown.button:GetFontString():SetTextColor(color.r, color.g, color.b)
  end
  
  local encounterName, description, encounterID, rootSectionID, link = EJ_GetEncounterInfo(currentEncounterID)
  local instanceName = EJ_GetInstanceInfo()
  local difficulty = GetDifficulty(link)
  local characterName = playerName
  local lastUpdate = time()
  
  BossList.frame.addEntry(currentInstanceID, difficulty, encounterID, characterName, desire, lastUpdate)
  SendUpdate(currentInstanceID, difficulty, encounterID, characterName, desire, lastUpdate)
end

function BossList_Menu(frame, level, menuList)
  local characterName = playerName
  local desire = 4
  if currentEncounterID ~= nil then
    local _, _, _, _, link = EJ_GetEncounterInfo(currentEncounterID)
    local entryDesire, _ = getEntry(BossList_entries, currentInstanceID, GetDifficulty(link), currentEncounterID, characterName)
    if entryDesire ~= nil then desire = entryDesire end
  end
  local info = UIDropDownMenu_CreateInfo()
  info.func = BossList_Menu_OnClick
  for idx, desireName in pairs(desires) do
    info.text, info.arg1, info.checked = desireName, idx, idx == desire
    UIDropDownMenu_AddButton(info)
  end
end

local function ToggleChoiceDropDown(button)
  PlaySound("igMainMenuOptionCheckBoxOn")
  currentEncounterID = button.encounterID
  dropDown.button = button
  ToggleDropDownMenu(1, nil, dropDown, "cursor", 3, -3)
end

local function DisplayInstance_Hook(instanceID, noButton)
  instanceID = instanceID or currentInstanceID
  --print("DisplayInstance" .. (instanceID or "nil"))
  if not EJ_InstanceIsRaid() then -- Only display if the instance is a raid
    return
  end
  
  local _G = _G
  local counter = 1
  local BossEntry = _G["EncounterJournalBossButton"..counter]
  local difficulty = GetDifficulty(BossEntry.link)
  while BossEntry ~= nil do
    local bossButton = BossEntry.bossButton
    if bossButton == nil then -- Create a button if it doesn't exist yet
      bossButton = CreateFrame("Button", "BossListbossButton", BossEntry, "EncounterDifficultyTemplate")
      
      bossButton:Raise()
      bossButton:SetScript("OnClick", ToggleChoiceDropDown)
      bossButton:SetPoint("RIGHT",BossEntry,-20,0)
      --bossButton:SetNormalTexture("Interface\\Buttons\\UI-GroupLoot-Dice-Up")
      BossEntry.bossButton = bossButton
    end
    
    local desire, _ = getEntry(BossList_entries, instanceID, difficulty, BossEntry.encounterID, playerName) or 4
    
    bossButton:SetText(desireIcons[desire])
    local color = desireColors[desire]
    if color ~= nil then
      bossButton:GetFontString():SetTextColor(color.r, color.g, color.b)
    end
    bossButton.encounterID = BossEntry.encounterID
    bossButton:Show()
    
    counter = counter + 1
    BossEntry = _G["EncounterJournalBossButton"..counter]
  end

  currentInstanceID = instanceID
end

function Init()
  dropDown = CreateFrame("Frame", "BossListContextMenu", UIParent, "UIDropDownMenuTemplate")
  UIDropDownMenu_Initialize(dropDown, BossList_Menu, "MENU")
    
  hooksecurefunc("EncounterJournal_DisplayInstance", DisplayInstance_Hook);
  hooksecurefunc("EJ_SetDifficulty", function() DisplayInstance_Hook() end);
	hooksecurefunc(EncounterJournalNavBarHomeButton,"myclick", HideButtons);
end

-- Register chat prefixes
AceComm:RegisterComm(FULL_MSG_PREFIX, UpdateHandler)
AceComm:RegisterComm(SMALL_MSG_PREFIX, UpdateHandler)
AceComm:RegisterComm(ADDON_PREFIX, MessageHandler)
frame:RegisterEvent("ADDON_LOADED")
frame:SetScript("OnEvent", EventHandler)


optionsTable = {
  type = "group",
  args = {
    minimap = {
      name = "Hide minimap button",
      desc = "Hides/shows the minimap button",
      width = "double",
      type = "toggle",
      set = function(info,val)
              if val then 
                Icon:Hide("BossList") 
                DB.char.minimap.hide = true
              else 
                Icon:Show("BossList")
                DB.char.minimap.hide = false
              end end,
      get = function(info) return DB.char.minimap.hide end
    },
    notNeeded = {
      name = "Hide \"Not Needed\" entries",
      desc = [[Hides/shows "Not Needed" in the BossList overview
      
Warning: decreases performance when opening/filtering the overview]],
      width = "double",
      type = "toggle",
      set = function(info,val)
              if val then 
                DB.global.hideNotNeeded = true
              else 
                DB.global.hideNotNeeded = false
              end 
              if BossList.frame.visible then
                BossList.frame:Show()
              end end,              
      get = function(info) return DB.global.hideNotNeeded end
    }
  }
}

LibStub("AceConfig-3.0"):RegisterOptionsTable("BossList", optionsTable);
LibStub("AceConfigDialog-3.0"):AddToBlizOptions("BossList", "BossList");

-- Register slash commands
SLASH_BOSSLIST1 = "/bosslist"
SLASH_BOSSLIST2 = "/bl"
SlashCmdList["BOSSLIST"] = function(msg)
  local arg1, arg2, arg3, arg4 = strsplit(" ", msg)
  if arg1 == "resync" then
    if arg2 ~= nil then
      RequestSync(arg2)
    else
      RequestUpdater()
    end
  else
    BossList.frame:Show()
  end
end